/*
	          Name: Pseudo-random buffer-encrypter/decrypter
	       Version: 3.0.0-alpha
	    Written by: TCH (tch@protonmail.com, http://oscomp.hu)
	  Contributors: kemi
	 Last modified: 2025.01.03.
	Released under: Public Domain

$VER: Pseudo-random buffer-encrypter/decrypter 3.0.0-alpha (03.01.2025)
*/
#ifndef _RC3_C
#define _RC3_C

/* Original concept, mode2 algorithm and C++ port by kemi. Original (Delphi) implementation, mode1 algorithm and C port by TCH. */

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "psrand.c"

#define RC3_EBADPTR	1

typedef struct
{
	psrandseed_t r;
	uint8_t t[256];
} rc_t;

void _rc3_gen_table3(rc_t *rc, uint32_t keyval)
{
	int i;
	uint8_t ri, tmp;

	spsrand(&rc->r, keyval);
	for (i = 0; i < 256; ++i)
	{
		ri = psrand(&rc->r) & 0xff;
		tmp = rc->t[i];
		rc->t[i] = rc->t[ri];
		rc->t[ri] = tmp;
	}
}

void _rc3_gen_table2(rc_t *rc, uint32_t keyval)
{
	int i;
	bool bt[256];
	uint8_t gb;

	for (i = 0; i < 256; ++i)
	{
		bt[i] = false;
	}
	spsrand(&rc->r, keyval);
	i = 0;
	while (i < 256)
	{
		gb = psrand(&rc->r) & 0xff;
		if (!bt[gb])
		{
			bt[gb] = true;
			rc->t[i++] = gb;
		}
	}
}

void (*_rc_gen_table[2])(rc_t *, uint32_t) = {&_rc3_gen_table3, &_rc3_gen_table2};

int rc3_crypt(rc_t *rc, bool mode2, bool gen2, size_t window_size, uint8_t *src, uint8_t *dst, size_t size, char *password)
{
	int i, lpw;
	uint32_t tv, rs;
	uint8_t c;
	size_t tp, wp;

	if ((rc == NULL) || (src == NULL) || (dst == NULL) || (password == NULL))
	{
		return RC3_EBADPTR;
	}

	mode2 = !!mode2;
	gen2 = !!gen2;

	if (!gen2)
	{
		for (i = 0; i < 256; ++i)
		{
			rc->t[i] = (uint8_t)i;
		}
	}

	tv = 0;
	lpw = strlen(password);
	for (i = 0; i < lpw; ++i)
	{
		tv += (uint8_t)password[i];
	}
	rs = tv / lpw;
	tv *= lpw;
	spsrand(&rc->r, tv);
	c = (uint8_t)password[lpw - 1];
	rs += psrandivu(&rc->r, 0, tv) + ((uint32_t)c * (uint32_t)c);

	tp = 0;
	wp = window_size;
	while (tp < size)
	{
		if (++wp >= window_size)
		{
			wp = 0;
			_rc_gen_table[gen2](rc, rs);
		}

		c = src[tp];
		if (mode2)
		{
			c = rc->t[c];
		}
		else
		{
			i = 0;
			while (i < 256)
			{
				if (c == rc->t[i])
				{
					
					c = (uint8_t)i;
					break;
				}
				++i;
			}
		}
		dst[tp++] = c;
	}
	return 0;
}

#endif
